@@ -1,13 +1,16 @@
-"""
-The `graders` module contains all the grader classes.
-"""
+"""The ``graders`` module contains all the grader classes."""
+
 import pprint
 import random
 import re
 from math import isclose, isnan
 
 import markdown
+import numpy as np
 import pandas as pd
+from PIL import Image
+from PIL.PngImagePlugin import PngImageFile
+from scipy import ndimage
 from sklearn.base import is_classifier, is_regressor
 from sklearn.exceptions import NotFittedError
 from sklearn.metrics import (
@@ -65,6 +68,7 @@ class BaseGrader(object):
             )
 
     def __repr__(self) -> str:
+        """Pretty dictionary representation of grader object."""
         rep_dict = {
             "points": self.points,
             "submission dtype": type(self.submission),
@@ -176,14 +180,15 @@ class BaseGrader(object):
             "passed": self.passed,
             "comment": comment,
         }
+
+        if hasattr(self, "diff_path"):
+            feedback_dict["diff"] = self.diff_path
+
         return feedback_dict
 
 
 class PythonGrader(BaseGrader):
-    """
-    Grader for evaluating `data types <https://docs.python.org/3/library/stdtypes.html>`_ from the Python standard library.
-
-    """
+    """Evaluate data types from the Python standard library."""
 
     def __init__(self, submission, answer, points=1):
         super().__init__(submission, answer, points)
@@ -326,7 +331,6 @@ class PythonGrader(BaseGrader):
         The value for key `b` doesn't match the expected result.
 
         """
-
         if not isinstance(self.submission, dict):
             raise TypeError(
                 f"grade_dict method can only be used with dict submissions, not {type(self.submission).__name__}."
@@ -396,10 +400,7 @@ class PythonGrader(BaseGrader):
 
 
 class PandasGrader(BaseGrader):
-    """
-    Grader for evaluating objects from `pandas <https://pandas.pydata.org/docs/index.html>`_.
-
-    """
+    """Grader for evaluating objects from `pandas <https://pandas.pydata.org/docs/index.html>`_. library."""
 
     def __init__(self, submission, answer, points=1):
         super().__init__(submission, answer, points)
@@ -407,9 +408,9 @@ class PandasGrader(BaseGrader):
     # https://tinyurl.com/y3sg2umv
     @staticmethod
     def _clean_assert_message(message: AssertionError) -> str:
-        """Helper function for making feedback student-friendly.
+        """Make feedback student-friendly.
 
-        Used by ``grade_df`` and ``grade_series``.
+        Helper function used by ``grade_df`` and ``grade_series``.
         """
         message = str(message)
 
@@ -508,7 +509,6 @@ class PandasGrader(BaseGrader):
         [submission]:  Int64Index([14, 16, 4], dtype='int64')
         [answer]: Int64Index([16, 14, 4], dtype='int64')
         """
-
         if not isinstance(self.submission, pd.DataFrame):
             raise TypeError(
                 f"grade_df method can only be used with DataFrames submissions, not {type(self.submission).__name__}."
@@ -682,10 +682,7 @@ class PandasGrader(BaseGrader):
 
 
 class SklearnGrader(BaseGrader):
-    """
-    Grader for evaluating objects from `sckit-learn <https://scikit-learn.org/stable/>`_.
-
-    """
+    """Grader for evaluating objects from `sckit-learn <https://scikit-learn.org/stable/>`_."""
 
     def __init__(self, submission, answer, points=1):
         super().__init__(submission, answer, points)
@@ -719,6 +716,10 @@ class SklearnGrader(BaseGrader):
             values must be identical. If ``tolerance=0.1``, values must be within 10%
             of each other (relative to the larger absolute value of the two).
 
+        return_bool : bool, default=False
+            Whether to return ``self.passed`` once grading is complete. You'll need this if
+            you want to design your own grading workflow beyond the default.
+
         Examples
         --------
         Let's create two linear models that use different scalers. We'll then fit only
@@ -763,7 +764,6 @@ class SklearnGrader(BaseGrader):
         "Step 1 in your model Pipeline doesn't match the expected result.
         Expected: `StandardScaler`. Received: `MinMaxScaler`."
         """
-
         # Is the answer model fitted?
         try:
             check_is_fitted(self.answer)
@@ -861,6 +861,87 @@ class SklearnGrader(BaseGrader):
         tolerance=0.0,
         return_bool=False,
     ):
+        """Evaluate model's performance using the model itself.
+
+        Parameters
+        ----------
+        X_test: pd.DataFrame
+            The test data feature matrix.
+
+        y_test: pd.Series
+            The test data target vector.
+
+        metric: str {"accuracy_score", "precision_score", "recall_score", "f1_score", "r2_score", "mean_absolute_error", "mean_squared_error"}
+            Metric to use when evaluating model performance.
+
+        round_to: int, default=3
+            Number of decimal places to round metric to before comparing
+            submission and answer model performance.
+
+        tolerance : int or float, default=0.0
+            What is the maximum allowed difference between submission and answer
+            model performance. If ``tolerance=0.0``, submission and answer metrics
+            must be identical. If ``tolerance=0.1``, values must be within 10%
+            of each other (relative to the larger absolute value of the two).
+
+        return_bool : bool, default=False
+            Whether to return ``self.passed`` once grading is complete. You'll need this if
+            you want to design your own grading workflow beyond the default.
+
+        Examples
+        --------
+        Let's start by creating a dataset, and splitting it into train and test.
+
+        >>> from sklearn.datasets import fetch_california_housing
+        >>> from sklearn.dummy import DummyRegressor
+        >>> from sklearn.model_selection import train_test_split
+        >>> X, y = fetch_california_housing(return_X_y=True, as_frame=True)
+        >>> X_train, X_test, y_train, y_test = train_test_split(
+        ...     X, y, test_size=0.2, random_state=42
+        ... )
+
+        Next, we create an answer and submission model, and put them into a grader.
+
+        >>> ans = DummyRegressor(strategy="constant", constant=2).fit(X_train, y_train)
+        >>> sub = DummyRegressor(strategy="constant", constant=1).fit(X_train, y_train)
+        >>> g = SklearnGrader(sub, ans)
+
+        Finally, we grade model performance looking at MSE.
+
+        >>> g.grade_model_performance(
+        ...     X_test=X_test,
+        ...     y_test=y_test,
+        ...     metric="mean_absolute_error",
+        ...     return_bool=True,
+        ... )
+        False
+        >>> print(g.comment)
+        Your model's mean absolute error is `1.141`. You can do better. Try to
+        beat `0.893`.
+
+        If we allow for some tolerance, the model will pass.
+
+        >>> g.grade_model_performance(
+        ...     X_test=X_test,
+        ...     y_test=y_test,
+        ...     metric="mean_absolute_error",
+        ...     tolerance=0.5,
+        ...     return_bool=True,
+        ... )
+        True
+        >>> print(g.comment)
+        Your model's mean absolute error is `1.141`. Very impressive.
+
+        """
+        if not isinstance(X_test, pd.DataFrame):
+            raise AttributeError(
+                f"X_test must be a DataFrame, not {type(X_test)}."
+            )
+
+        if not isinstance(y_test, pd.Series):
+            raise AttributeError(
+                f"y_test must be a Series, not {type(y_test)}."
+            )
         # Whether submission model outperforms answer
         sub_beats_ans = False
 
@@ -985,6 +1066,62 @@ class SklearnGrader(BaseGrader):
         tolerance=0.0,
         return_bool=False,
     ):
+        """Evaluate model's performance using the predicitons itself.
+
+        Submission and answer must be ``pd.Series``.
+
+        Parameters
+        ----------
+        metric: str {"accuracy_score", "precision_score", "recall_score", "f1_score", "r2_score", "mean_absolute_error", "mean_squared_error"}
+            Metric to use when evaluating predicitons.
+
+        threshold: float
+            Score that predictions must beat.
+
+        round_to: int, default=3
+            Number of decimal places to round metric to before comparing
+            predictions score to threshold.
+
+        tolerance : int or float, default=0.0
+            What is the maximum allowed difference between prediction score
+            and threshold. If ``tolerance=0.0``, predictions must beat threshold.
+            If ``tolerance=0.1``, values must be within 10% of each other (relative to
+            the larger absolute value of the two).
+
+        Examples
+        --------
+        Let's start by creating a dataset.
+
+        >>> from sklearn.datasets import fetch_california_housing
+        >>> from sklearn.dummy import DummyRegressor
+        >>> from sklearn.model_selection import train_test_split
+        >>> import pandas as pd
+        >>> from grading_tools.graders import SklearnGrader
+        >>> X, y = fetch_california_housing(return_X_y=True, as_frame=True)
+        >>> X_train, X_test, y_train, y_test = train_test_split(
+        ...     X, y, test_size=0.2, random_state=42
+        ... )
+
+        Next, we'll train a model and generate a Series of predictions
+
+        >>> model = DummyRegressor(strategy="constant", constant=2).fit(X_train, y_train)
+        >>> y_pred = pd.Series(model.predict(X_test))
+
+        Finally, we'll put the predictions and true values in a grader.
+
+        >>> g = SklearnGrader(y_pred, y_test)
+
+        The submission will pass as long as the metric score beats the threshold.
+
+        >>> g.grade_model_predictions(
+        ... metric="mean_absolute_error",
+        ... threshold=0.9,
+        ... return_bool=True,
+        ... )
+        >>> print(g.comment)
+        Your model's mean absolute error is `0.893`. Boom! You got it.
+
+        """
         # Whether submission model outperforms answer
         sub_beats_threshold = False
 
@@ -1085,3 +1222,148 @@ class SklearnGrader(BaseGrader):
             return self.passed
         else:
             return None
+
+
+class PlotGrader(BaseGrader):
+    """Grader for evaluating images."""
+
+    def __init__(self, submission, answer, points=1):
+        super().__init__(submission, answer, points)
+
+    def grade_plot_images(
+        self,
+        threshold=0.0,
+        return_diff=True,
+        diff_path="./diff.png",
+        highlight_size=50,
+        return_bool=False,
+    ):
+        """Compare two images.
+
+        Evaluates how similar two images are by calculating the root mean square error
+        between their pixels.
+
+        Inspired by Matplotlib's `compare_images <https://matplotlib.org/devdocs/api/testing_api.html#matplotlib.testing.compare.compare_images>`_.
+
+        Parameters
+        ----------
+        threshold: float, default=0.0
+            The RMSE under which the submitted image needs to score in order to still
+            be considered correct.
+
+        return_diff: bool, default=True
+            Whether to generate a file highlighting difference between submission and
+            answer. If ``True``, file is saved to ``diff_path``.
+
+        diff_path: str, default="./diff.png"
+            Location to which the diff image will be saved.
+
+        highlight_size: int, default=50
+            How large the overlayed highlights in the diff image should be.
+
+        return_bool : bool, default=False
+            Whether to return ``self.passed`` once grading is complete. You'll need this if
+            you want to design your own grading workflow beyond the default.
+
+        """
+        # Check that submission type is correct
+        if not isinstance(self.submission, PngImageFile):
+            raise TypeError(
+                "grade_plot_images method can only be used on `PIL.Image` objects, "
+                f"not {type(self.submission).__name__} objects."
+            )
+
+        # The submission needs to be the same size as answer
+        if self.submission.size != self.answer.size:
+            self.update_comment(
+                "Your submission plot is not the same size as the expected result."
+                "Make sure your plot has the fig size specified in the instructions"
+                "and that you're saving the plot with the correct `dpi`."
+            )
+            if return_bool:
+                return self.passed
+            else:
+                return None
+
+        # Turn images into B&W ndarrays
+        sub_bw_array = np.asarray(self.submission.convert("1")).astype(
+            np.int16
+        )
+        ans_bw_array = np.asarray(self.answer.convert("1")).astype(np.int16)
+
+        # If arrays are equal, passed
+        if threshold <= 0:
+            if np.array_equal(sub_bw_array, ans_bw_array):
+                self.add_to_score()
+                if return_bool:
+                    return self.passed
+                else:
+                    return None
+
+        # Calculate the per-pixel errors, then compute the root mean square error.
+        rmse = np.sqrt(
+            ((sub_bw_array - ans_bw_array).astype(float) ** 2).mean()
+        )
+
+        # If RMSE below threshold, passed
+        if rmse < threshold:
+            self.add_to_score()
+            if return_bool:
+                return self.passed
+            else:
+                return None
+
+        # If RMSE is above threshold and you want a diff image
+        if rmse > threshold and return_diff:
+            # If it's not below threshold, make diff image
+            # Step 1: Create diff array
+            diff_bw_array = np.abs(ans_bw_array - sub_bw_array)
+
+            # Step 2. Create diff B&W highlights array
+            s = np.ones((highlight_size, highlight_size))
+            diff_hl_bw = ndimage.morphology.binary_dilation(
+                diff_bw_array, structure=s
+            ).astype(int)
+
+            # Step 3. Create RGB version of diff
+            # 3.1. Start with empty array. The `4` refers RBG + alpha channels
+            diff_hl_rgb = np.zeros(list(diff_hl_bw.shape) + [4])
+            # 3.2. Add red pixels
+            diff_hl_rgb[:, :, 0] = diff_hl_bw * 255
+            # 3.3. Add alpha
+            diff_hl_rgb[:, :, -1] = 125
+            # 3.4. Add white pixels
+            diff_hl_rgb[
+                np.where((diff_hl_rgb == [0, 0, 0, 125]).all(axis=2))
+            ] = [
+                255,
+                255,
+                255,
+                0,
+            ]
+            diff_hl_rgb = diff_hl_rgb.astype(np.uint8)
+
+            # Step 4: Create diff image
+            diff_hl_img = Image.fromarray(diff_hl_rgb)
+            self.diff = self.submission.copy()
+            self.diff.paste(diff_hl_img, (0, 0), mask=diff_hl_img)
+
+            # Step 5: Save diff image
+            self.diff.save(diff_path, format="png")
+            self.diff_path = diff_path
+
+            # Step 6: Update feedback
+            self.update_comment(
+                "Your submission doesn't match the expected result. "
+                "Check the image below to see where your plot differs from the answer."
+            )
+
+        if rmse > threshold and not return_diff:
+            self.update_comment(
+                "Your submission doesn't match the expected result."
+            )
+
+        if return_bool:
+            return self.passed
+        else:
+            return None
